#!/usr/bin/env python
#coding=utf-8
"""
Created on Sun Aug 04 10:05:26 2013

@author: Deyuan Zhang and Bin Liu
This file contains the utilities of leanring the optimal fusion of kernel 
matrix using Kernel Target Alignment described in the paper. 
"""

from numpy import *
import sys
import cvxopt
from cvxopt import solvers
from gistutil import getMat, saveMat, getLabel

def centerMat(mattrain):
    """the function center the kernel matrix into zero mean and unit variance.
    numpy.array, int = centerMat(numpy.array)"""
    matmean = tile(mean(mattrain), mattrain.shape)
    meanx = ones((mattrain.shape[0],1))*reshape(mean(mattrain, axis=0),(1,-1))
    meany = reshape(mean(mattrain, axis=1),(-1,1))
    meany = meany*ones((1,mattrain.shape[0]))
    mattraincenter = mattrain - meanx - meany + matmean
    norm = mean(diag(mattraincenter))
    mattraincenter /= norm
    return mattraincenter, norm


def getWeightKTA(mats, labels, fname):
    """getWeightKTA implements the Kernel Target Alignment technique of the 
    paper, especially the equation of the paper.
    mats is a list of numpy array, and each array corresponds to a kernel matrix
    labels is a list of the labels of the training set, +1 or -1
    fname is used to save the learned weight to file, and fname is the filename of the file.
    It returns the """
    matnum = len(mats)
    matst = []
    norm = zeros(matnum)
    for i,mat in enumerate(mats):
        matt,normt = centerMat(mat)
        matst.append(matt)
        norm[i] = normt
    mats = matst
    matLargeLen = len(labels)**2
    matD = zeros((matLargeLen, matnum))
    for i,mat in enumerate(mats):
        matD[:,i]=mat.flatten()
    labels = array(labels)
    
    G = dot(labels.reshape(-1,1),labels.reshape(1,-1))
    G, nG = centerMat(G)
    Q = dot(G.flatten(), matD)
    Q = Q/sum(Q)
    
    P = dot(matD.T, matD)       
    P = cvxopt.matrix(P)
    q = cvxopt.matrix( 0.0, (matnum, 1) )
    A = cvxopt.matrix(Q, (1,matnum))
    b = cvxopt.matrix( 1.0 )
    G = cvxopt.matrix( diag(array([-1.0]*matnum)) )
    h = cvxopt.matrix( 0.0, (matnum, 1) )
    
    sol=solvers.qp(P, q, G, h, A, b)
    x = array(sol['x'])
    x /= sum(x)
    savetxt(fname, x.reshape(-1,1))
    matWeight= (array(sol['x']).flatten()/norm).tolist()
    print 'weight',matWeight
    return matWeight

def getSelfMat(fnames, weight):
    """given the kernels to be combined and the corresponding weights, this fuction produce the fusioned kernel(both training and testing) into GIST format.
    the input fnames is a list which contains 8 filenames, the format is [train_kernel1, test_kernel1, train_kernel2, test_kernel2, train_kernel3, test_kernel3, fusion_train_kernel, fusion_test_kernel]
    the weight is a list that contains 3 values, indicate the corresponding weight of each kernel"""
    if len(fnames) != 8:
        print 'self mat wrong'
    trainmat = zeros_like(loadtxt(fnames[0], skiprows = 1, usecols = (0,1)))
    
    testmat = zeros_like(loadtxt(fnames[1], skiprows = 1, usecols = (0,1)))
    print trainmat.shape, testmat.shape
    for (ind,f_ind) in enumerate([0,2,4]):
        mat=loadtxt(fnames[f_ind], skiprows = 1, usecols = (0,1))
        trainmat[:,1]+=mat[:,1]*weight[ind]
        trainmat[:,0]=mat[:,0]
    for (ind,f_ind) in enumerate([1,3,5]):
        mat=loadtxt(fnames[f_ind], skiprows = 1, usecols = (0,1))
        testmat[:,1]+=mat[:,1]*weight[ind]
        testmat[:,0]=mat[:,0]

    f = open(fnames[6],'w')
    f.write('seq\t<<value\n')
    mat = trainmat
    for i in range(mat.shape[0]):
        f.write("%d\t%e\n"%(mat[i,0],mat[i,1]))
    f.close()
    f = open(fnames[7],'w')
    mat = testmat
    f.write('seq\t<<value\n')
    for i in range(mat.shape[0]):
        f.write("%d\t%e\n"%(mat[i,0],mat[i,1]))
    f.close()

if __name__ == '__main__':
    if len(sys.argv) != 20:
        print 'check your argument'
        print len(sys.argv)
        sys.exit()
    #read the 
    mtype = sys.argv[1]
    flname = sys.argv[2]    
    ftrainname1 = sys.argv[3]
    ftestname1 = sys.argv[4]    
    ftrainname2 = sys.argv[5]
    ftestname2 = sys.argv[6]    
    ftrainname3 = sys.argv[7]
    ftestname3 = sys.argv[8]
    ftrainname = sys.argv[9]
    ftestname = sys.argv[10]
    
    trainlabel = getLabel(flname)

    mattrain = []    
    mattest = []
    trainmat1 = getMat(ftrainname1)
    mattrain.append(trainmat1)
    testmat1 = getMat(ftestname1)
    mattest.append(testmat1)
    
    trainmat2 = getMat(ftrainname2)
    mattrain.append(trainmat2)
    testmat2 = getMat(ftestname2)
    mattest.append(testmat2)
    
    trainmat3 = getMat(ftrainname3)
    mattrain.append(trainmat3)
    testmat3 = getMat(ftestname3)
    mattest.append(testmat3)    

    if mtype == "KTA":
        print 'kernel combination type: KTA'
        weight = getWeightKTA(mattrain, trainlabel, sys.argv[-1])

    print weight
    
    matfinaltrain = zeros_like(mattrain[0])
    matfinaltest = zeros_like(mattest[0])
    for mattraintemp,mattesttemp,weighttemp in zip(mattrain, mattest, weight):
        matfinaltrain += mattraintemp*weighttemp
        matfinaltest += mattesttemp*weighttemp
    saveMat(ftrainname, matfinaltrain)
    saveMat(ftestname, matfinaltest)
    
    getSelfMat(sys.argv[11:19],weight)
